{
   uFileHandlers
   Provides mechanisms for registering and calling multiple file handlers.
   Copyright (C) 2007. Dejan Boras

   This file is part of uFile.

   uFile is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   uFile is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with uFile. If not, see <http://www.gnu.org/licenses/>.

   Started On:    22.06.2007.
   Last Update:   19.03.2011.
}

{$MODE OBJFPC}{$H+}{$I-}
UNIT uFileHandlers;

INTERFACE

   USES uStd, {$IFNDEF NO_LOGGING}uLog, {$ENDIF}uError, SysUtils, uFile;

CONST
   {error codes}
   fheGENERAL                    = $0100;
   fheFILE_EXIST_INVALID         = $0101;
   fheHANDLER_NOT_FOUND          = $0102;

   {constants}
   fhcGuessExtension: boolean    = false;

TYPE
   {file handler data, passed to a file handle proc}
   fhPData = ^fhTData;
   fhTData = record
      fn: string;
   end;

   {a procedure which is called to handle files}
   fhTHandleProc = procedure(data: pointer; const fhdata: fhTData);

   {a handler, which is responsible for handling files}
   fhPHandler = ^fhTHandler;
   fhTHandler = record
      Name: string;
      Handle: fhTHandleProc;

      DoNotOpenFile: boolean; {instructs not to open the file before the handler is called}
      next: fhPHandler;
   end;

   {an extension, associated with a handler}
   fhPExtension = ^fhTExtension;
   fhTExtension = record
      ext: string;
      handler: fhPHandler;

      next: fhPExtension;
   end;

   {magic number}
   fhTMagicNo = record
      Size: longint;
      mNo: array[0..15] of char;
   end;

   {handler information}
   fhPHandlerInfo = ^fhTHandlerInfo;
   fhTHandlerInfo = record
      nExtensions, nHandlers: longint;

      hStart, hEnd: fhPHandler;
      eStart, eEnd: fhPExtension;

      defaultHandler: fhPHandler;
      prevHandlerExt: fhPExtension;      
   end;   

   fhTFileExistFunc = function(const fn: string): fileint;

   fhPFindData = ^fhTFindData;
   fhTFindData = record
      handler: fhPHandler;
      ext: fhPExtension;
   end;

VAR
   fhError: longint;

{error handling}
procedure fheRaise(err: longint);
procedure fhErrorReset();

{initialize a handler info record}
procedure fhInitHandlerInfoRecord(var hinfo: fhTHandlerInfo); inline;

{extension and handler registration}
procedure fhRegisterExt(ext: fhPExtension; var hinfo: fhTHandlerInfo);
procedure fhRegisterHandler(ldr: fhPHandler; var hinfo: fhTHandlerInfo);

{load a file}
procedure fhLoad(const FileName: string; var hinfo: fhTHandlerInfo; var fl: file; data: pointer);
procedure fhLoad(var ldr: fhTHandler; data: pointer; const fhdata: fhTData);

{write a file}
procedure fhWrite(const FileName: string; var hinfo: fhTHandlerInfo; var fl: file; data: pointer);
procedure fhWrite(var w: fhTHandler; data: pointer; const fhdata: fhTData);

{sets the file exist routine to be used with fhFindHandler}
procedure fhSetFileExistFunc(fe: fhTFileExistFunc);

{finds a valid handler for a file}
procedure fhFindHandler(const filename: string; var hinfo: fhTHandlerInfo; var fd: fhTFindData);

{logging}
procedure fhLogExtensions(const hinfo: fhTHandlerInfo);

IMPLEMENTATION

VAR
   fExist: fhTFileExistFunc;

{report an error}
procedure fheRaise(err: longint);
begin
   fhError := err;
end;

procedure fhErrorReset();
begin
   fhError := 0;
end;

{handler initialize}
procedure fhInitHandlerInfoRecord(var hinfo: fhTHandlerInfo); inline;
begin
   Zero(hinfo, sizeof(fhTHandlerInfo));
end;

{register extension}
procedure fhRegisterExt(ext: fhPExtension; var hinfo: fhTHandlerInfo);
begin
   inc(hinfo.nExtensions); ext^.Next := nil;
   if(hinfo.eStart = nil) then begin
      hinfo.eStart := ext;
      hinfo.eEnd := ext;
   end else begin
      hinfo.eEnd^.Next := ext;
      hinfo.eEnd := ext;
   end;
end;

{register an handler}
procedure fhRegisterHandler(ldr: fhPHandler; var hinfo: fhTHandlerInfo);
begin
   inc(hinfo.nHandlers); ldr^.Next := nil;
   if(hinfo.hStart = nil) then begin
      hinfo.hStart := ldr;
      hinfo.hEnd := ldr;
   end else begin
      hinfo.hEnd^.Next := ldr;
      hinfo.hEnd := ldr;
   end;

   if(hinfo.defaultHandler = nil) then hinfo.defaultHandler := ldr;
end;

{the main load routine}
procedure fhLoad(const FileName: string; var hinfo: fhTHandlerInfo; var fl: file; data: pointer);
var
   fName: string;
   fd: fhTFindData;
   fhdata: fhTData;

begin
   fhErrorReset();
   fName := FileName;

   {$PUSH}{$HINTS OFF}fhFindHandler(FileName, hinfo, fd);{$POP}

   if(fd.handler <> nil) then begin
      {open the file if allowed by the handler}
      if(not fd.handler^.DoNotOpenFile) then begin
         {initialize the file}
         Assign(fl, fName);
         Reset(fl, 1);
         if(ioerror <> 0) then begin fheRaise(eIO); exit; end;
      end;

      fhdata.fn := fName;
      fd.handler^.Handle(data, fhdata);

      if(not fd.handler^.DoNotOpenFile) then begin
         Close(fl);
         if(ioerror <> 0) then fheRaise(eIO);
      end;
   end else begin
      fheRaise(fheHANDLER_NOT_FOUND);
      {$IFNDEF NO_LOGGING}
      logWriteln('uFileHandlers > No handler could be found for the file.');{$ENDIF}
   end;
end;

procedure fhLoad(var ldr: fhTHandler; data: pointer; const fhdata: fhTData);
begin
   if(ldr.Handle <> nil) then ldr.Handle(data, fhdata);
end;

procedure fhWrite(const FileName: string; var hinfo: fhTHandlerInfo; var fl: file; data: pointer);
var
   fName: string;
   fd: fhTFindData;
   fhdata: fhTData;

begin
   fhErrorReset();
   fName := FileName;

   {$PUSH}{$HINTS OFF}fhFindHandler(FileName, hinfo, fd);{$POP}

   if(fd.handler <> nil) then begin
      {open the file if allowed by the handler}
      if(not fd.handler^.DoNotOpenFile) then begin
         {initialize the file}
         Assign(fl, fName);
         Rewrite(fl, 1);
         if(ioerror <> 0) then begin fheRaise(eIO); exit; end;
      end;

      fhdata.fn := fName;
      fd.handler^.Handle(data, fhdata);

      if(not fd.handler^.DoNotOpenFile) then begin
         Close(fl);
         if(ioerror <> 0) then fheRaise(eIO);
      end;
   end else begin
      fheRaise(fheHANDLER_NOT_FOUND);
      {$IFNDEF NO_LOGGING}
      logWriteln('uFileHandlers > No handler could be found for the file.');{$ENDIF}
      exit;
   end;
end;

procedure fhWrite(var w: fhTHandler; data: pointer; const fhdata: fhTData);
begin
   if(w.Handle <> nil) then w.Handle(data, fhdata);
end;


{DISPOSE}

function fhFileExist(const fn: string): fileint;
begin
   fhFileExist := fExist(fn);
end;

{sets the file exist routine to be used with fhFindHandler}
procedure fhSetFileExistFunc(fe: fhTFileExistFunc);
begin
   if(fe <> nil) then fExist := fe else fExist := @fhFileExist;
end;

{finds a valid handler for a file}
procedure fhFindHandler(const filename: string; var hinfo: fhTHandlerInfo; var fd: fhTFindData);
var
   fName, fn: string;
   ext: string; {extension of the file}
   cur: fhPExtension;

procedure DetermineHandler();
var
   cur: fhPExtension;

begin
   cur := hinfo.eStart;
   repeat
      if(cur^.ext = ext) then begin fd.ext := cur; exit; end;
      cur := cur^.Next;
   until (cur = nil);
end;

begin
   fErrorReset();
   fd.ext := nil; fd.handler := nil;

   {if there are no extensions registered then we can't load anything :(}
   if(hinfo.nExtensions > 0) then begin
      fName := filename;

      {determine extension}
      ext := LowerCase(ExtractFileExt(fName));
      if(ext <> '') then fhcGuessExtension := false;

      {determine which handler to use and call it}
      if(fhcGuessExtension = false) then begin {determine the handler by the extension}
         if(ext <> '') then begin {if there is an extension}
            if(hinfo.prevHandlerExt <> nil) then begin
               if(hinfo.prevHandlerExt^.ext = ext) then fd.ext := hinfo.prevHandlerExt
               else DetermineHandler();
            end else DetermineHandler();

            if(fd.ext <> nil) then begin
               {call the default handler}
               hinfo.prevHandlerExt := fd.ext; fd.handler := fd.ext^.handler;
            end;
         end else begin {if no extension then call the default handler}
            if(hinfo.defaultHandler <> nil) then fd.handler := hinfo.defaultHandler;
         end;
      {check out all extensions and see if the file exists, if it does use the handler
      associated with the extension.}
      end else begin
         if(fExist <> nil) then begin
            fd.handler := hinfo.defaultHandler;
            {check all extensions}
            cur := hinfo.eStart;
            repeat
               fn := fName+cur^.ext;
               if(fExist(fn) >= 0) then begin {file found}
                  fName := fn; fd.handler := cur^.Handler; fd.ext := cur; break;
               end;
               cur := cur^.Next;
            until (cur = nil);
         end else fheRaise(fheFILE_EXIST_INVALID);
      end;
   end;
end;

procedure fhLogExtensions(const hinfo: fhTHandlerInfo);
var
   cur: fhPExtension;

begin
   cur := hinfo.eStart;
   if(cur <> nil) then repeat
      {$IFNDEF NO_LOGGING}
      logWriteln(cur^.ext+'('+cur^.handler^.Name+')');
      {$ENDIF}
      cur := cur^.Next;
   until (cur = nil);
end;

INITIALIZATION
   fExist := @fhFileExist;
END.
